@@ -51,6 +51,7 @@
 #include "utils/rel.h"
 #include "utils/syscache.h"
 #include "utils/timestamp.h"
+#include "utils/varlena.h"
 
 /* Default passord encryption */
 #define Password_encryption = PASSWORD_TYPE_SCRAM_SHA_256;
@@ -211,6 +212,7 @@ static char *username_not_contain = NULL;
 static char *username_contain = NULL;
 static bool username_contain_password = true;
 static bool username_ignore_case = false;
+static char *username_whitelist = NULL;
 
 /* Password flags*/
 static int password_min_length = 1;
@@ -239,6 +241,9 @@ static int password_reuse_interval = 0;
 char *str_to_sha256(const char *str, const char *salt);
 #endif
 
+bool check_whitelist(char **newval, void **extra, GucSource source);
+bool is_in_whitelist(char *username);
+
 static char *to_nlower(const char *str, size_t max) {
   char *lower_str;
   int i = 0;
@@ -471,6 +476,71 @@ username_check(const char *username, const char *password)
 	free(tmp_not_contains);
 }
 
+/* We just check that the list is valid, no username existing check */
+bool
+check_whitelist(char **newval, void **extra, GucSource source)
+{
+	char       *rawstring;
+	List       *elemlist;
+
+	/* Need a modifiable copy of string */
+	rawstring = pstrdup(*newval);
+	/* Parse string into list of identifiers */
+	if (!SplitIdentifierString(rawstring, ',', &elemlist))
+	{
+		/* syntax error in list */
+		GUC_check_errdetail("List syntax is invalid.");
+		pfree(rawstring);
+		list_free(elemlist);
+		return false;
+	}
+
+	pfree(rawstring);
+	list_free(elemlist);
+
+	return true;
+}
+
+/* check if the username is in the whitelist */
+bool
+is_in_whitelist(char *username)
+{
+	char       rawstring[NAMEDATALEN];
+	List       *elemlist;
+	ListCell   *l;
+
+	Assert(username != NULL);
+
+	/* Need a modifiable copy of string */
+	strcpy(rawstring, username_whitelist);
+	/* Parse string into list of identifiers */
+	if (!SplitIdentifierString(rawstring, ',', &elemlist))
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
+				errmsg("%s username list is invalid: %s",
+				     "credcheck.password_min_length", username_whitelist)));
+		list_free(elemlist);
+		return false;
+	}
+
+        foreach(l, elemlist)
+        {
+                char       *tok = (char *) lfirst(l);
+
+		/* the username is in the list */
+                if (pg_strcasecmp(tok, username) == 0)
+		{
+                        list_free(elemlist);
+                        return true;
+                }
+        }
+
+	list_free(elemlist);
+
+	return false;
+}
+
 static void password_check(const char *username, const char *password)
 {
 
@@ -484,6 +554,9 @@ static void password_check(const char *username, const char *password)
 	char *tmp_contains = NULL;
 	char *tmp_not_contains = NULL;
 
+	Assert(username != NULL);
+	Assert(password != NULL);
+
 	/* checks has to be done by ignoring case */
 	if (password_ignore_case)
 	{
@@ -1029,11 +1102,13 @@ check_password_reuse(const char *username, const char *password)
 	char         *encrypted_password;
 	HASH_SEQ_STATUS hash_seq;
 
-	if (password_reuse_history == 0 && password_reuse_interval == 0)
+	Assert(username != NULL);
+
+	if (password == NULL)
 		return false;
 
-	Assert(username != NULL);
-	Assert(password != NULL);
+	if (password_reuse_history == 0 && password_reuse_interval == 0)
+		return false;
 
 	/* Safety check... */
 	if (!pgph || !pgph_hash)
@@ -1157,17 +1232,23 @@ check_password(const char *username, const char *password,
 #ifdef USE_CRACKLIB
 			const char *reason;
 #endif
+			if (is_in_whitelist((char *)username))
+				break;
+
 			statement_has_password = true;
 			username_check(username, password);
-			password_check(username, password);
+			if (password != NULL)
+			{
+				password_check(username, password);
 #ifdef USE_CRACKLIB
-			/* call cracklib to check password */
-			if ((reason = FascistCheck(password, CRACKLIB_DICTPATH)))
-				ereport(ERROR,
-						(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-						 errmsg("password is easily cracked"),
-						 errdetail_log("cracklib diagnostic: %s", reason)));
+				/* call cracklib to check password */
+				if ((reason = FascistCheck(password, CRACKLIB_DICTPATH)))
+					ereport(ERROR,
+							(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+							 errmsg("password is easily cracked"),
+							 errdetail_log("cracklib diagnostic: %s", reason)));
 #endif
+			}
 			break;
 		}
 		default:
@@ -1220,6 +1301,11 @@ _PG_init(void)
 				NULL, &encrypted_password_allowed, false, PGC_SUSET, 0,
 				NULL, NULL, NULL);
 
+	DefineCustomStringVariable(
+				"credcheck.whitelist",
+				gettext_noop("comma separated list of username to exclude from password policy check"), NULL,
+				&username_whitelist, "", PGC_SUSET, 0, check_whitelist, NULL, NULL);
+
 
 #if PG_VERSION_NUM < 150000
         /*
@@ -1288,7 +1374,9 @@ cc_ProcessUtility(PEL_PROCESSUTILITY_PROTO)
 			/* We only take care of user renaming */
 			if (stmt->renameType == OBJECT_ROLE && stmt->newname != NULL)
 			{
-
+				if (is_in_whitelist(stmt->newname) || is_in_whitelist(stmt->subname))
+					break;
+			
 				/* check the validity of the username */
 				username_check(stmt->newname, NULL);
 
@@ -1306,36 +1394,50 @@ cc_ProcessUtility(PEL_PROCESSUTILITY_PROTO)
 			ListCell      *option;
 			char          *password;
 			bool           save_password = false;
+			DefElem    *dpassword = NULL;
+			DefElem    *dvalidUntil = NULL;
+
+			if (is_in_whitelist(stmt->role->rolename))
+				break;
 
 			/* Extract options from the statement node tree */
 			foreach(option, stmt->options)
 			{
 				DefElem    *defel = (DefElem *) lfirst(option);
 
-#if PG_VERSION_NUM >= 120000
 				if (strcmp(defel->defname, "password") == 0)
 				{
-					statement_has_password = true;
-					password = strVal(defel->arg);
-					save_password = check_password_reuse(stmt->role->rolename, password);
+					dpassword = defel;
 				}
+				else if (strcmp(defel->defname, "validUntil") == 0)
+				{
+					dvalidUntil = defel;
+				}
+			}
+
+#if PG_VERSION_NUM >= 120000
+			if (dpassword && dpassword->arg)
+			{
+				statement_has_password = true;
+				password = strVal(dpassword->arg);
+				save_password = check_password_reuse(stmt->role->rolename, password);
+			}
 #endif
-				if (password_valid_until > 0 && strcmp(defel->defname, "validUntil") == 0)
-				{
-					int valid_until = check_valid_until(strVal(defel->arg));
-					if (valid_until < password_valid_until)
-						ereport(ERROR,
-							(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
-								errmsg(gettext_noop("the VALID UNTIL option must have a date older than %d days"), password_valid_until)));
-				}
-				if (password_valid_max > 0 && strcmp(defel->defname, "validUntil") == 0)
-				{
-					int valid_max = check_valid_until(strVal(defel->arg));
-					if (valid_max > password_valid_max)
-						ereport(ERROR,
-							(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
-								errmsg(gettext_noop("the VALID UNTIL option must NOT have a date beyond %d days"), password_valid_max)));
-				}
+			if (dvalidUntil && dvalidUntil->arg && password_valid_until > 0)
+			{
+				int valid_until = check_valid_until(strVal(dvalidUntil->arg));
+				if (valid_until < password_valid_until)
+					ereport(ERROR,
+						(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
+							errmsg(gettext_noop("the VALID UNTIL option must have a date older than %d days"), password_valid_until)));
+			}
+			if (dvalidUntil && dvalidUntil->arg && password_valid_max > 0)
+			{
+				int valid_max = check_valid_until(strVal(dvalidUntil->arg));
+				if (valid_max > password_valid_max)
+					ereport(ERROR,
+						(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
+							errmsg(gettext_noop("the VALID UNTIL option must NOT have a date beyond %d days"), password_valid_max)));
 			}
 
 			/* The password can be saved into the history */
@@ -1353,6 +1455,11 @@ cc_ProcessUtility(PEL_PROCESSUTILITY_PROTO)
 			bool            has_valid_until = false; 
 			bool            save_password = false;
 			char           *password;
+			DefElem    *dpassword = NULL;
+			DefElem    *dvalidUntil = NULL;
+
+			if (is_in_whitelist(stmt->role))
+				break;
 
 			/* check the validity of the username */
 			username_check(stmt->role, NULL);
@@ -1362,25 +1469,35 @@ cc_ProcessUtility(PEL_PROCESSUTILITY_PROTO)
 			{
 				DefElem    *defel = (DefElem *) lfirst(option);
 
-#if PG_VERSION_NUM >= 120000
 				if (strcmp(defel->defname, "password") == 0)
 				{
-					statement_has_password = true;
-					password = strVal(defel->arg);
-					save_password = check_password_reuse(stmt->role, password);
+					dpassword = defel;
 				}
-#endif
-				if (password_valid_until > 0 && strcmp(defel->defname, "validUntil") == 0)
+				else if (strcmp(defel->defname, "validUntil") == 0)
 				{
-					valid_until = check_valid_until(strVal(defel->arg));
-					has_valid_until = true;
-				}
-				if (password_valid_max > 0 && strcmp(defel->defname, "validUntil") == 0)
-				{
-					valid_max = check_valid_until(strVal(defel->arg));
-					has_valid_until = true;
+					dvalidUntil = defel;
 				}
 			}
+
+#if PG_VERSION_NUM >= 120000
+			if (dpassword && dpassword->arg)
+			{
+				statement_has_password = true;
+				password = strVal(dpassword->arg);
+				save_password = check_password_reuse(stmt->role, password);
+			}
+#endif
+			if (dvalidUntil && dvalidUntil->arg && password_valid_until > 0)
+			{
+				valid_until = check_valid_until(strVal(dvalidUntil->arg));
+				has_valid_until = true;
+			}
+			if (dvalidUntil && dvalidUntil->arg && password_valid_max > 0)
+			{
+				valid_max = check_valid_until(strVal(dvalidUntil->arg));
+				has_valid_until = true;
+			}
+
 			/* check that a VALID UNTIL option is present */
 			if ( !has_valid_until && (password_valid_until > 0 || password_valid_max > 0) )
 				ereport(ERROR,
